package de.schlegel11.jfxanimation;

import javafx.animation.Timeline;
import javafx.scene.Node;

import java.util.*;
import java.util.function.Function;
import java.util.stream.Collectors;

/**
 * Class which represents a general {@link JFXAnimationTemplate}. <br>
 * This class is responsible for providing methods where it's possible to build a CSS like
 * structured animation.
 *
 * @author Marcel Schlegel (schlegel11)
 * @version 1.0
 * @since 2018-09-18
 */
public class JFXAnimationTemplate<N> implements JFXTemplateConfig<N>, JFXTemplateBuilder<N> {

  private static final String NO_KEY_FOUND_MESSAGE =
      "No animation objects with key \"%s\" found.\n"
          + "Please check your build method where the "
          + "namedAnimationObjects are defined or your "
          + "withAnimationObject methods where the keys are accessed.";

  private final Set<Double> percents = new HashSet<>();
  private final Map<
          Double,
          List<
              Function<
                  JFXAnimationTemplateAction.InitBuilder<N>,
                  JFXAnimationTemplateAction.Builder<?, ?>>>>
      creatorValueBuilderFunctions = new HashMap<>();
  private final Class<N> animationObjectType;
  private Map<String, Collection<Object>> animationObjects;
  private Function<JFXAnimationTemplateConfig.Builder, JFXAnimationTemplateConfig.Builder>
      creatorConfigBuilderFunction;
  private boolean clearPercents;

  private JFXAnimationTemplate(Class<N> animationObjectType) {
    this.animationObjectType = animationObjectType;
  }

  /**
   * Create a {@link JFXTemplateProcess} with a specific default animation type.<br>
   * The default animation objects with this type are set later in the {@link
   * JFXTemplateBuilder#build(Object)} method.<br>
   * These objects are generally used in {@link JFXTemplateAction#action(Function)} methods.
   *
   * @param animationObjectType a specific animation object type.
   * @param <N> the specific type.
   * @return a {@link JFXTemplateProcess} instance.
   */
  public static <N> JFXTemplateProcess<N> create(Class<N> animationObjectType) {
    return new JFXAnimationTemplate<>(animationObjectType);
  }

  /**
   * Same method as {@link #create(Class)} but with the default type {@link Node}. This type is a
   * general default type in a {@link JFXAnimationTemplate}.
   *
   * @see #create(Class)
   * @return a {@link JFXTemplateProcess} instance.
   */
  public static JFXTemplateProcess<Node> create() {
    return create(Node.class);
  }

  public Map<Double, List<JFXAnimationTemplateAction<?, ?>>> buildAndGetAnimationValues() {

    Map<Double, List<JFXAnimationTemplateAction<?, ?>>> animationValueMap = new HashMap<>();
    creatorValueBuilderFunctions.forEach(
        (percent, animationValueBuilderFunctions) -> {
          List<JFXAnimationTemplateAction<?, ?>> animationValues =
              animationValueBuilderFunctions
                  .stream()
                  .flatMap(
                      builderFunction ->
                          builderFunction
                              .apply(JFXAnimationTemplateAction.builder(animationObjectType))
                              .buildActions(this::getAnimationObjectsWithKeys))
                  .collect(Collectors.toList());
          animationValueMap.put(percent, animationValues);
        });
    return animationValueMap;
  }

  private List<Object> getAnimationObjectsWithKeys(Collection<String> keys) {
    List<Object> animationObjectList = new ArrayList<>();
    for (String key : keys) {
      Collection<?> animationObjectsPerKey = animationObjects.get(key);
      if (animationObjectsPerKey == null) {
        throw new NoSuchElementException(String.format(NO_KEY_FOUND_MESSAGE, key));
      }
      animationObjectList.addAll(animationObjectsPerKey);
    }
    return animationObjectList;
  }

  public JFXAnimationTemplateConfig buildAndGetTemplateConfig() {
    return creatorConfigBuilderFunction.apply(JFXAnimationTemplateConfig.builder()).build();
  }

  @Override
  public JFXTemplateAction<N> percent(double first, double... rest) {
    if (clearPercents) {
      percents.clear();
      clearPercents = false;
    }
    // Clamp value between 0 and 100.
    percents.add(Math.max(0, Math.min(100, first)));
    creatorValueBuilderFunctions.put(first, new ArrayList<>());

    for (double percent : rest) {
      percents.add(Math.max(0, Math.min(100, percent)));
      creatorValueBuilderFunctions.put(percent, new ArrayList<>());
    }
    return this;
  }

  @Override
  public JFXTemplateAction<N> from() {
    return percent(0);
  }

  @Override
  public JFXTemplateAction<N> to() {
    return percent(100);
  }

  @Override
  public JFXTemplateConfig<N> action(
      Function<JFXAnimationTemplateAction.InitBuilder<N>, JFXAnimationTemplateAction.Builder<?, ?>>
          valueBuilderFunction) {
    for (Double percent : percents) {
      creatorValueBuilderFunctions.get(percent).add(valueBuilderFunction);
    }
    clearPercents = true;
    return this;
  }

  @Override
  public JFXTemplateConfig<N> action(
      JFXAnimationTemplateAction.Builder<?, ?> animationValueBuilder) {
    return action(builder -> animationValueBuilder);
  }

  @Override
  public JFXTemplateBuilder<N> config(
      Function<JFXAnimationTemplateConfig.Builder, JFXAnimationTemplateConfig.Builder>
          configBuilderFunction) {
    creatorConfigBuilderFunction = configBuilderFunction;
    return this;
  }

  @Override
  public JFXTemplateBuilder<N> config(JFXAnimationTemplateConfig.Builder configBuilder) {
    return config(builder -> configBuilder);
  }

  @Override
  public <B> B build(
      Function<JFXAnimationTemplate<N>, B> builderFunction,
      Function<JFXAnimationObjectMapBuilder<N>, JFXAnimationObjectMapBuilder<N>>
          mapBuilderFunction) {
    animationObjects =
        mapBuilderFunction.apply(JFXAnimationObjectMapBuilder.builder()).getAnimationObjects();
    // Provide a null value as default animation object if it's absent.
    animationObjects.putIfAbsent(
        JFXAnimationObjectMapBuilder.DEFAULT_ANIMATION_OBJECT_NAME,
        Collections.singletonList(null));
    return builderFunction.apply(this);
  }

  @Override
  public <B> B build(
      Function<JFXAnimationTemplate<N>, B> builderFunction, N defaultAnimationObject) {
    return build(builderFunction, b -> b.defaultObject(defaultAnimationObject));
  }

  @Override
  public Timeline build(
      Function<JFXAnimationObjectMapBuilder<N>, JFXAnimationObjectMapBuilder<N>>
          mapBuilderFunction) {
    return build(JFXAnimationTemplates::buildTimeline, mapBuilderFunction);
  }

  @Override
  public Timeline build(N defaultAnimationObject) {
    return build(b -> b.defaultObject(defaultAnimationObject));
  }

  @Override
  public Timeline build() {
    return build(Function.identity());
  }
}
